from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime
import pandas as pd
from rich.console import Console
from rich.text import Text
import subprocess
from textual import on
from textual.app import App, ComposeResult
from textual.containers import Grid, Horizontal, Vertical
from textual_plotext import PlotextPlot
from textual.reactive import reactive
from textual.screen import ModalScreen, Screen
from textual.validation import Number
from textual.widget import Widget
from textual.widgets import Button, Input, Header, Footer, Label, Tabs, DataTable, TabbedContent, TabPane, Static, Markdown, Pretty, Select

from terka.domain import _commands, events, models
# from terka.domain.task import Task
# from terka.domain.project import Project
# from terka.domain.sprint import Sprint
from terka.service_layer import services, views
from terka.service_layer.formatter import Formatter


class Comment(Widget):
    value = reactive("text")

    def render(self) -> str:
        return f"Comment: {self.value}"


class TerkaTask(App):
    BINDINGS = [("q", "quit", "Quit")]
    CSS_PATH = "vertical_layout.css"

    def __init__(self,
                 entity,
                 project,
                 is_completed: bool = False,
                 history=None,
                 commentaries=None):
        super().__init__()
        self.entity = entity
        self.project = project
        self.is_completed = True if entity.status.name in (
            "DONE", "DELETED") else False
        self.history = history
        self.commentaries = commentaries
        self.is_overdue = datetime.today().date(
        ) > entity.due_date if entity.due_date and entity.status.name not in (
            "DONE", "DELETED") else False

    def compose(self) -> ComposeResult:
        yield Header()
        task_text = f"[bold]Task {self.entity.id}: {self.entity.name}[/bold]"
        if self.is_completed:
            yield Static(task_text, classes="header_completed", id="header")
        elif self.is_overdue:
            yield Static(task_text, classes="header_overdue", id="header")
        else:
            yield Static(task_text, classes="header_simple", id="header")
        yield Static(f"Project: [bold]{self.project}[/bold]", classes="transp")
        yield Static(f"Status: [bold]{self.entity.status.name}[/bold]",
                     classes="transp")
        yield Static(f"Priority: [bold]{self.entity.priority.name}[/bold]",
                     classes="transp")
        created_days_ago = (datetime.now() - self.entity.creation_date).days
        creation_message = f"{created_days_ago} days ago" if created_days_ago > 1 else "today"
        yield Static(f"Created {creation_message}", classes="transp")
        if self.is_completed:
            completion_date = self.entity.completion_date
            if completion_date:
                completed_days_ago = (datetime.now() - completion_date).days
                completion_message = f"{completed_days_ago} days ago" if completed_days_ago > 1 else "today"
                yield Static(
                    f"Completed {completion_message} ({completion_date.strftime('%Y-%m-%d')})",
                    classes="transp")
            else:
                yield Static("Completion date unknown", classes="transp")
        else:
            yield Static(f"Due date: {self.entity.due_date}", classes="transp")
        if sprints := self.entity.sprints:
            sprint_id = ",".join(str(s.sprint) for s in sprints)
            story_points = sprints[-1].story_points
            yield Static(
                f"Sprints: [bold]{sprint_id}[/bold], "
                f"SP: [bold]{story_points}[/bold], "
                f"T: [bold]{Formatter.format_time_spent(self.entity.total_time_spent)}[/bold]",
                classes="transp")
        else:
            yield Static(f"Not in sprint", classes="transp")
        if epics := self.entity.epics:
            epic_ids = ",".join([str(e.epic) for e in epics])
            yield Static(f"Epics: [bold]{epic_ids}[/bold]", classes="transp")
        else:
            yield Static(f"Not in epic", classes="transp")
        if stories := self.entity.stories:
            story_ids = ",".join([str(s.story) for s in stories])
            yield Static(f"Story: [bold]{story_ids}[/bold]", classes="transp")
        else:
            yield Static(f"Not in story", classes="transp")
        if tags := self.entity.tags:
            tags = ",".join(
                t.base_tag.text for t in tags
                if not t.base_tag.text.startswith(("epic", "story", "sprint")))
            yield Static(f"Tags: [bold]{tags}[/bold]", classes="transp")
        else:
            yield Static(f"No tags", classes="transp")
        if collaborators := self.entity.collaborators:
            collaborators = ",".join(c.users.name for c in collaborators)
            yield Static(f"Collaborators: [bold]{collaborators}[/bold]",
                         classes="transp")
        else:
            yield Static(f"No collaborators", classes="transp")
        description_message = self.entity.description or ""
        if description_message:
            yield Static(
                f"[italic]Description...\n[/italic]{description_message}",
                classes="body",
                id="desc")
        else:
            yield Static("[italic]No description...[/italic]",
                         classes="body",
                         id="desc")
        if self.commentaries:
            comms = [
                f"[italic]{comment.date.date()}[/italic]: {comment.text}"
                for comment in self.commentaries
            ]
            comms_message = "\n".join(comms)
            yield Static(f"[italic]Comments...\n\n[/italic]{comms_message}",
                         classes="body",
                         id="history")
        else:
            yield Static("", classes="body", id="history")
        # TODO: implement during next release
        # yield Input(placeholder="Add a comment", classes="body", id="comment")
        # yield Pretty("")

    @on(Input.Submitted)
    def submit_input(self, event: Input.Submitted) -> None:
        self.query_one(Pretty).update("Comment added")
        input = self.query_one(Input)
        input.value = ""


class TerkaProject(App):

    CSS = """
    Tabs {
        dock: top;
    }
    QuitScreen {
        align: center middle;
    }
    TaskComplete {
        align: center top;
    }
    #dialog {
        grid-size: 2;
        grid-gutter: 1 2;
        grid-rows: 1fr 3;
        padding: 0 1;
        width: 60;
        height: 20;
        border: thick $background 80%;
        background: $surface;
    }

    #question {
        column-span: 2;
        height: 1fr;
        width: 1fr;
        content-align: center middle;
    }

    Button {
        width: 100%;
    }
    .plotext {
        width: 100%;
        height: 20;
    }
    .header {
        margin:1 1;
        width: 100%;
        height: 5%;
        background: $panel;
        border: tall $primary;
        content-align: center middle;
    }
    """

    BINDINGS = [
        ("b", "backlog", "Backlog"),
        ("t", "tasks", "Tasks"),
        ("e", "epics", "Epics"),
        ("s", "stories", "Stories"),
        ("n", "notes", "Notes"),
        ("o", "overview", "Overview"),
        ("T", "time", "Time"),
        ("q", "quit", "Quit"),
        ("E", "task_edit", "Edit"),
        ("r", "refresh", "Refresh"),
        ("d", "task_complete", "Done"),
        ("X", "task_delete", "Delete"),
    ]

    def __init__(self, repo, config, project_id, bus) -> None:
        super().__init__()
        self.repo = repo
        self.config = config
        self.bus = bus
        self.tasks = list()
        self.selected_task = None
        self.project_id = project_id
        self.entity = self.get_entity()

    def get_entity(self):
        return self.repo.get(models.project.Project, self.project_id)

    def action_refresh(self):
        self.entity = self.get_entity()

    def on_mount(self) -> None:
        self.title = f"Project: {self.entity.name}"
        self.sub_title = f'Workspace: {self.config.get("workspace")}'

    def action_task_complete(self) -> None:
        self.push_screen(TaskComplete(), self.task_complete_callback)

    def action_task_edit(self) -> None:
        task_edit = TaskEdit()
        self.push_screen(task_edit, self.task_complete_callback)

    def task_complete_callback(self, result: TaskCompletionInfo):
        self.bus.handle(
            _commands.CompleteTask(id=self.selected_task,
                                   comment=result.comment,
                                   hours=result.hours))
        self.notify(f"Task: {self.selected_task} is completed!")

    def action_task_delete(self) -> None:
        self.push_screen(TaskDelete(), self.task_delete_callback)

    def task_delete_callback(self, result: TaskCompletionInfo):
        self.bus.handle(
            _commands.DeleteTask(id=self.selected_task,
                                 comment=result.comment,
                                 hours=result.hours))
        self.notify(f"Task: {self.selected_task} is deleted!",
                    severity="warning")

    def compose(self) -> ComposeResult:
        yield Header()
        yield Static(f"{self.entity.name}", classes="header")
        with TabbedContent(initial="tasks"):
            with TabPane("Backlog", id="backlog"):
                table = DataTable(id="backlog")
                for column in ("id", "name", "priority", "due_date",
                               "created_at", "tags", "collaborators",
                               "time_spent"):
                    table.add_column(column, key=column)
                for task in sorted(self.entity.tasks,
                                   key=lambda x: x.id,
                                   reverse=True):
                    if task.status.name == "BACKLOG":
                        if tags := task.tags:
                            tags_text = ",".join(
                                [tag.base_tag.text for tag in list(tags)])
                        else:
                            tags_text = ""
                        if collaborators := task.collaborators:
                            collaborators_texts = sorted([
                                collaborator.users.name
                                for collaborator in list(collaborators)
                                if collaborator.users
                            ])
                            collaborator_string = ",".join(collaborators_texts)
                        else:
                            collaborator_string = ""
                        if task.is_overdue:
                            task_id = f"[red]{task.id}[/red]"
                        elif task.is_stale:
                            task_id = f"[yellow]{task.id}[/yellow]"
                        else:
                            task_id = str(task.id)
                        table.add_row(task_id,
                                      task.name,
                                      task.priority.name,
                                      task.due_date,
                                      task.creation_date.strftime("%Y-%m-%d"),
                                      tags_text,
                                      collaborator_string,
                                      Formatter.format_time_spent(
                                          task.total_time_spent),
                                      key=task.id)
                yield table
            with TabPane("Open Tasks", id="tasks"):
                table = DataTable(id="tasks")
                for column in ("id", "name", "status", "priority", "due_date",
                               "tags", "collaborators", "time_spent"):
                    table.add_column(column, key=column)
                for task in sorted(self.entity.tasks,
                                   key=lambda x: x.status.name,
                                   reverse=True):
                    # self.tasks.append(task)
                    if task.status.name in ("TODO", "IN_PROGRESS", "REVIEW"):
                        if tags := task.tags:
                            tags_text = ",".join(
                                [tag.base_tag.text for tag in list(tags)])
                        else:
                            tags_text = ""
                        if collaborators := task.collaborators:
                            collaborators_texts = sorted([
                                collaborator.users.name
                                for collaborator in list(collaborators)
                                if collaborator.users
                            ])
                            collaborator_string = ",".join(collaborators_texts)
                        else:
                            collaborator_string = ""
                        if task.is_overdue:
                            task_id = f"[red]{task.id}[/red]"
                        elif task.is_stale:
                            task_id = f"[yellow]{task.id}[/yellow]"
                        else:
                            task_id = str(task.id)
                        table.add_row(task_id,
                                      task.name,
                                      task.status.name,
                                      task.priority.name,
                                      task.due_date,
                                      tags_text,
                                      collaborator_string,
                                      Formatter.format_time_spent(
                                          task.total_time_spent),
                                      key=task.id)
                yield table
            with TabPane("Completed Tasks", id="completed_tasks"):
                table = DataTable(id="completed_tasks")
                for column in ("id", "name", "status", "priority",
                               "completion_date", "tags", "collaborators",
                               "time_spent"):
                    table.add_column(column, key=column)
                completed_tasks = [
                    e for e in self.entity.tasks if e.completion_date
                ]
                for task in sorted(completed_tasks,
                                   key=lambda x: x.completion_date,
                                   reverse=True):
                    if tags := task.tags:
                        tags_text = ",".join(
                            [tag.base_tag.text for tag in list(tags)])
                    else:
                        tags_text = ""
                    if collaborators := task.collaborators:
                        collaborators_texts = sorted([
                            collaborator.users.name
                            for collaborator in list(collaborators)
                            if collaborator.users
                        ])
                        collaborator_string = ",".join(collaborators_texts)
                    else:
                        collaborator_string = ""
                    table.add_row(
                        str(task.id), task.name, task.status.name,
                        task.priority.name,
                        task.completion_date.strftime("%Y-%m-%d"), tags_text,
                        collaborator_string,
                        Formatter.format_time_spent(task.total_time_spent))
                yield table
            with TabPane("Epics", id="epics"):
                table = DataTable(id="epics")
                table.add_columns("id", "name", "description", "status",
                                  "tasks")
                for epic in sorted(self.entity.epics,
                                   key=lambda x: len(x.tasks),
                                   reverse=True):
                    table.add_row(str(epic.id), shorten_text(epic.name),
                                  shorten_text(epic.description),
                                  epic.status.name, str(len(epic.tasks)))
                yield table
            with TabPane("Stories", id="stories"):
                table = DataTable(id="stories")
                table.add_columns("id", "name", "description", "status",
                                  "tasks")
                for story in self.entity.stories:
                    table.add_row(str(story.id), shorten_text(story.name),
                                  shorten_text(story.description),
                                  story.status.name, str(len(story.tasks)))
                yield table
            with TabPane("Notes", id="notes"):
                table = DataTable(id="notes")
                table.add_columns("id", "text")
                for task in self.entity.notes:
                    table.add_row(str(task.id), task.name)
                yield table
            with TabPane("Time", id="time"):
                plotext = PlotextPlot(classes="plotext")
                plt = plotext.plt
                n_days = 14
                time = self.entity.daily_time_entries_hours(last_n_days=n_days)
                plt.date_form('Y-m-d')
                plt.title(
                    f"Time tracker - {round(sum(time.values()))} hours spent"
                    f" for the last {n_days} days")
                plt.bar(time.keys(), time.values())
                yield plotext
            with TabPane("Overview", id="overview"):
                collaborators = self.entity.task_collaborators
                sorted_collaborators = ""
                for name, value in sorted(collaborators.items(),
                                          key=lambda x: x[1],
                                          reverse=True):
                    sorted_collaborators += f"  * {name}: {Formatter.format_time_spent(value)} \n"
                yield Markdown(f"""
# Project details:
* Repo: {self.entity.description}
* Time spend: {Formatter.format_time_spent(self.entity.total_time_spent)}
* Collaborators:
{sorted_collaborators}
                """)

        yield Footer()

    def action_epics(self) -> None:
        self.query_one(TabbedContent).active = "epics"

    def action_backlog(self) -> None:
        self.query_one(TabbedContent).active = "backlog"

    def action_tasks(self) -> None:
        self.query_one(TabbedContent).active = "tasks"

    def action_stories(self) -> None:
        self.query_one(TabbedContent).active = "stories"

    def action_overview(self) -> None:
        self.query_one(TabbedContent).active = "overview"

    def action_notes(self) -> None:
        self.query_one(TabbedContent).active = "notes"

    def action_time(self) -> None:
        self.query_one(TabbedContent).active = "time"

    def action_done(self):
        self.selected_task = 0

    def on_data_table_cell_highlighted(self, event: DataTable.CellHighlighted):
        self.selected_task = event.cell_key.row_key.value

    def on_data_table_cell_selected(self, event: DataTable.CellSelected):
        self.selected_task = event.cell_key.row_key.value
        # subprocess.run(["/home/am/envs/terka-textual/bin/terka", "done", "task", "1911"])
        # exit(str(event.cell_key.row_key.value))


@dataclass
class TaskCompletionInfo:
    comment: str | None = None
    hours: int | None = None


class TaskStatusEdit(ModalScreen[str]):

    def compose(self) -> ComposeResult:
        yield Grid(
            Label(f"Change status", id="question"),
            Select(((line, line) for line in [
                "BACKLOG", "TODO", "IN_PROGRESS", "REVIEW", "DONE", "DELETED"
            ]),
                   prompt="status",
                   id="status"),
            Button("Confim", id="yes"),
            Button("Cancel", id="no"),
            id="dialog",
        )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "yes":
            status = self.query_one("#status", Select)
            self.dismiss(status.value)
        else:
            self.app.pop_screen()

    @on(Input.Submitted)
    def submit_input(self, event: Input.Submitted) -> None:
        exit(self)


class TaskComplete(ModalScreen[str]):

    def compose(self) -> ComposeResult:
        yield Grid(
            Label(f"Complete task", id="question"),
            Input(placeholder="Add comment", id="comment"),
            Input(placeholder="Add time spent",
                  validators=[Number()],
                  id="hours"),
            Button("Yes", id="yes"),
            Button("No", id="no"),
            id="dialog",
        )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "yes":
            comment = self.query_one("#comment", Input)
            hours = self.query_one("#hours", Input)
            self.dismiss(TaskCompletionInfo(comment.value, hours.value))
        else:
            self.app.pop_screen()

    @on(Input.Submitted)
    def submit_input(self, event: Input.Submitted) -> None:
        exit(self)


class TaskDelete(ModalScreen[str]):

    def compose(self) -> ComposeResult:
        yield Grid(
            Label(f"Delete task", id="question"),
            Input(placeholder="Add comment", id="comment"),
            Input(placeholder="Add time spent",
                  validators=[Number()],
                  id="hours"),
            Button("Yes", id="yes"),
            Button("No", id="no"),
            id="dialog",
        )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "yes":
            comment = self.query_one("#comment", Input)
            hours = self.query_one("#hours", Input)
            self.dismiss(TaskCompletionInfo(comment.value, hours.value))
        else:
            self.app.pop_screen()

    @on(Input.Submitted)
    def submit_input(self, event: Input.Submitted) -> None:
        exit(self)


class TaskEdit(ModalScreen[str]):

    def compose(self) -> ComposeResult:
        yield Grid(
            Label(f"Edit task", id="question"),
            Input(placeholder="", id="name"),
            Input(placeholder="", id="description"),
            Select(((line, line) for line in ["TODO", "IN_PROGRESS"]),
                   prompt="status",
                   id="status"),
            Select(
                ((line, line) for line in ["LOW", "NORMAL", "HIGH", "URGENT"]),
                prompt="priority",
                id="priority"),
            Input(placeholder="Add comment", id="comment"),
            Input(placeholder="Add time spent",
                  validators=[Number()],
                  id="hours"),
            Button("Yes", id="yes"),
            Button("No", id="no"),
            id="dialog",
        )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "yes":
            comment = self.query_one("#comment", Input)
            hours = self.query_one("#hours", Input)
            exit(f"Comment: {comment.value}, hours: {hours.value}")
            self.app.exit()
        else:
            self.app.pop_screen()

    @on(Input.Submitted)
    def submit_input(self, event: Input.Submitted) -> None:
        exit(self)


class TerkaSprint(App):

    CSS = """
    Tabs {
        dock: top;
    }
    QuitScreen {
        align: center middle;
    }
    TaskComplete {
        align: center top;
    }
    #dialog {
        grid-size: 2;
        grid-gutter: 1 2;
        grid-rows: 1fr 3;
        padding: 0 1;
        width: 60;
        height: 20;
        border: thick $background 80%;
        background: $surface;
    }

    #question {
        column-span: 2;
        height: 1fr;
        width: 1fr;
        content-align: center middle;
    }

    Button {
        width: 100%;
    }
    Screen {
        align: center top;
    }
    .plotext {
        width: 100%;
        height: 20;
    }
    .header {
        margin:1 1;
        width: 100%;
        height: 5%;
        background: $panel;
        border: tall $primary;
        content-align: center middle;
    }
    """

    BINDINGS = [("t", "tasks", "Tasks"), ("n", "notes", "Notes"),
                ("o", "overview", "Overview"), ("q", "quit", "Quit"),
                ("P", "sort_by_project", "Sort by Project"),
                ("S", "sort_by_status", "Sort by Status"),
                ("E", "task_edit", "Edit"), ("r", "refresh", "Refresh"),
                ("d", "task_complete", "Done"), ("X", "task_delete", "Delete"),
                ("U", "task_update_context", "Update"), ("T", "time", "Time")]

    current_sorts: set = set()

    def __init__(self, repo, config, sprint_id, bus) -> None:
        super().__init__()
        self.repo = repo
        self.config = config
        self.bus = bus
        self.selected_task = None
        self.sprint_id = sprint_id
        self.entity = self.get_entity()
        self.tasks = self.get_tasks()

    def get_entity(self):
        return self.repo.get_by_id(models.sprint.Sprint, self.sprint_id)

    def get_tasks(self):
        for sprint_task in self.entity.tasks:
            if sprint_task.tasks.status.name not in ("DONE", "DELETED"):
                yield sprint_task.tasks

    def action_refresh(self):
        self.entity = self.get_entity()
        previous_tasks = list(self.tasks)
        self.tasks = list(self.get_tasks())
        new_tasks = list(self.tasks)
        exit(f"{len(new_tasks)}, {len(previous_tasks)}")
        self.notify("Refreshing data...")

    def sort_reverse(self, sort_type: str):
        """Determine if `sort_type` is ascending or descending."""
        reverse = sort_type in self.current_sorts
        if reverse:
            self.current_sorts.remove(sort_type)
        else:
            self.current_sorts.add(sort_type)
        return reverse

    def action_task_update_context(self) -> None:
        self.push_screen(TaskStatusEdit(), self.task_update_context_callback)

    def task_update_context_callback(self, result: str):
        self.bus.handle(
            _commands.UpdateTask(id=self.selected_task,
                                 status=models.task.TaskStatus[result]))
        self.notify(f"Task: {self.selected_task} status updated to {result}!")

    def action_task_complete(self) -> None:
        self.push_screen(TaskComplete(), self.task_complete_callback)

    def task_complete_callback(self, result: TaskCompletionInfo):
        self.bus.handle(
            _commands.CompleteTask(id=self.selected_task,
                                   comment=result.comment,
                                   hours=result.hours))
        self.notify(f"Task: {self.selected_task} is completed!")

    def action_task_delete(self) -> None:
        self.push_screen(TaskDelete(), self.task_delete_callback)

    def task_delete_callback(self, result: TaskCompletionInfo):
        self.bus.handle(
            _commands.DeleteTask(id=self.selected_task,
                                 comment=result.comment,
                                 hours=result.hours,
                                 entity_type="sprint",
                                 entity_id=self.entity.id))
        self.notify(
            f"Task: {self.selected_task} is deleted from the current sprint!",
            severity="warning")

    def action_sort_by_project(self) -> None:
        table = self.query_one("#tasks_table", DataTable)
        table.sort(
            "project",
            reverse=self.sort_reverse("project"),
        )

    def action_sort_by_status(self) -> None:
        table = self.query_one("#tasks_table", DataTable)
        table.sort(
            "status",
            reverse=self.sort_reverse("status"),
        )

    def on_mount(self) -> None:
        self.title = "Sprint"
        self.sub_title = f'Workspace: {self.config.get("workspace")}'

    def compose(self) -> ComposeResult:
        yield Header()
        yield Static(
            f"[bold]Sprint {self.entity.id}[/bold]: \n{self.entity.goal}",
            classes="header")
        with TabbedContent(initial="tasks"):
            with TabPane("Tasks", id="tasks"):
                table = DataTable(id="tasks_table")
                for column in ("id", "name", "status", "priority",
                               "story_points", "project", "due_date", "tags",
                               "collaborators", "time_spent"):
                    table.add_column(column, key=column)
                for sprint_task in sorted(self.entity.tasks,
                                          key=lambda x: x.tasks.status.value,
                                          reverse=True):
                    task = sprint_task.tasks
                    # self.tasks.append(task)
                    if tags := task.tags:
                        tags_text = ",".join([
                            tag.base_tag.text for tag in list(tags)
                            if not tag.base_tag.text.startswith("sprint")
                        ])
                    else:
                        tags_text = ""
                    if collaborators := task.collaborators:
                        collaborators_texts = sorted([
                            collaborator.users.name
                            for collaborator in list(collaborators)
                            if collaborator.users
                        ])
                        collaborator_string = ",".join(collaborators_texts)
                    else:
                        collaborator_string = ""
                    if task.is_overdue:
                        task_id = f"[red]{task.id}[/red]"
                    elif task.is_stale:
                        task_id = f"[yellow]{task.id}[/yellow]"
                    else:
                        task_id = str(task.id)
                    project = services.lookup_project_name(
                        task.project, self.repo)
                    if task.status.name not in ("DONE", "DELETED"):
                        table.add_row(task_id,
                                      task.name,
                                      task.status.name,
                                      task.priority.name,
                                      str(sprint_task.story_points),
                                      project,
                                      task.due_date,
                                      tags_text,
                                      collaborator_string,
                                      Formatter.format_time_spent(
                                          task.total_time_spent),
                                      key=task.id)
                yield table
            with TabPane("Completed Tasks", id="completed_tasks"):
                table = DataTable(id="completed_tasks")
                for column in ("id", "name", "project", "completion_date",
                               "tags", "collaborators", "story_points",
                               "time_spent"):
                    table.add_column(column, key=column)
                for sprint_task in sorted(self.entity.tasks,
                                          key=lambda x: x.id,
                                          reverse=True):
                    task = sprint_task.tasks
                    if tags := task.tags:
                        tags_text = ",".join([
                            tag.base_tag.text for tag in list(tags)
                            if not tag.base_tag.text.startswith("sprint")
                        ])
                    else:
                        tags_text = ""
                    if collaborators := task.collaborators:
                        collaborators_texts = sorted([
                            collaborator.users.name
                            for collaborator in list(collaborators)
                            if collaborator.users
                        ])
                        collaborator_string = ",".join(collaborators_texts)
                    else:
                        collaborator_string = ""
                    project = services.lookup_project_name(
                        task.project, self.repo)
                    if task.status.name in ("DONE", "DELETED"):
                        table.add_row(
                            str(task.id),
                            task.name,
                            project,
                            task.completion_date.strftime("%Y-%m-%d")
                            if task.completion_date else "",
                            tags_text,
                            collaborator_string,
                            Formatter.format_time_spent(
                                round(sprint_task.story_points * 60)),
                            Formatter.format_time_spent(task.total_time_spent),
                            key=task.id)
                yield table
            with TabPane("Notes", id="notes"):
                table = DataTable(id="notes")
                table.add_columns("id", "text")
                for task in self.entity.notes:
                    table.add_row(str(task.id), task.name)
                yield table
            with TabPane("Time", id="time"):
                plotext = PlotextPlot(classes="plotext")
                plt = plotext.plt
                if not (sprint_time := self.entity.daily_time_entries_hours()):
                    yield Static("No time tracked")
                else:
                    workspace = services.get_workplace_by_name(
                        self.config.get("workspace"), self.repo)
                    all_workspace_time = workspace.daily_time_entries_hours(
                        start_date=self.entity.start_date,
                        end_date=self.entity.end_date)
                    non_sprint_times = [
                        all_times - sprint_times
                        for all_times, sprint_times in zip(
                            all_workspace_time.values(), sprint_time.values())
                    ]

                    plt.date_form('Y-m-d')
                    plt.title(
                        f"Time tracker - {Formatter.format_time_spent(self.entity.total_time_spent)} spent"
                    )
                    plt.stacked_bar(sprint_time.keys(),
                                    [sprint_time.values(), non_sprint_times],
                                    label=["sprint", "non-sprint"])
                    yield plotext
            with TabPane("Overview", id="overview"):
                collaborators = self.entity.collaborators
                sorted_collaborators = ""
                for name, value in sorted(collaborators.items(),
                                          key=lambda x: x[1],
                                          reverse=True):
                    sorted_collaborators += f"  * {name}: {Formatter.format_time_spent(value)} \n"
                yield Markdown(f"""
# Sprint details:
* Period: {self.entity.start_date} - {self.entity.end_date}
* Open tasks: {len(self.entity.open_tasks)} ({len(self.entity.tasks)})
* Pct Completed: {round(self.entity.pct_completed, 2) :.0%}
* Velocity: {self.entity.velocity}
* Time spend: {Formatter.format_time_spent(self.entity.total_time_spent)}
* Utilization: {round(self.entity.utilization, 2) :.0%}
* Collaborators:
{sorted_collaborators}
                """)

        yield Footer()

    def action_tasks(self) -> None:
        self.query_one(TabbedContent).active = "tasks"

    def action_time(self) -> None:
        self.query_one(TabbedContent).active = "time"

    def action_overview(self) -> None:
        self.query_one(TabbedContent).active = "overview"

    def action_notes(self) -> None:
        self.query_one(TabbedContent).active = "notes"

    def on_data_table_cell_highlighted(self, event: DataTable.CellHighlighted):
        self.selected_task = event.cell_key.row_key.value

    def on_data_table_cell_selected(self, event: DataTable.CellSelected):
        self.selected_task = event.cell_key.row_key.value

    def on_data_table_cell_selected(self, event: DataTable.CellHighlighted):
        self.selected_task = event.cell_key.row_key.value


# TODO: implement
# class TerkaProjectScreen(TerkaProject):
#     ...

# class TerkaSprintScreen(TerkaSprint):
#     ...

# class Terka(App):
#     BINDINGS = [
#         ("1", "switch_mode('sprint')", "Sprint"),
#         ("2", "switch_mode('projects')", "Projects"),
#     ]
#     MODES = {"project": TerkaProjectScreen, "sprint": TerkaSprintScreen}

#     def on_mount(self) -> None:
#         self.switch_mode("project")


def shorten_text(text: str | None, limit: int = 80) -> str | None:
    if text and len(text) > limit:
        text = text[:limit]
    return text


def get_times(dates: list[str],
              time_entries: list[dict[str, str | int]]) -> list[float]:
    times: list[float] = []
    for date in dates:
        value_for_date = 0
        for entry in time_entries:
            if entry.get("date") == date:
                value_for_date = entry.get("time_spent_hours") / 60
                break
        times.append(value_for_date)
    return times



from typing import Type
from collections import defaultdict
from dataclasses import asdict
from datetime import datetime
import functools
import logging

from terka.adapters import publisher, printer
from terka.domain import _commands, events, models
from terka.service_layer import exceptions, templates, unit_of_work

COMMAND_HANDLERS = {}
EVENT_HANDLERS = defaultdict(list)


def register(cmd=None, event=None):

    def fn(func):

        @functools.wraps(func)
        def inner_function(cmd, handler, context):
            return func(cmd, handler, context)

        if cmd:
            COMMAND_HANDLERS[cmd] = inner_function
        elif event:
            EVENT_HANDLERS[event].append(inner_function)
        return inner_function

    return fn


class Handler:

    def __init__(
        self,
        uow: unit_of_work.AbstractUnitOfWork,
        publisher: publisher.BasePublisher = None,
        printer: printer.Printer = printer.Printer()
    ) -> None:
        self.uow = uow
        self.publisher = publisher
        self.printer = printer


class SprintCommandHandlers:

    @register(cmd=_commands.CreateSprint)
    def create(cmd: _commands.CreateSprint,
               handler: Handler,
               context: dict = {}) -> None:
        if not cmd:
            cmd, context = templates.create_command_from_editor(
                models.sprint.Sprint, type(cmd))
        with handler.uow as uow:
            # TODO: Add code for checking existing sprints
            new_sprint = models.sprint.Sprint(**asdict(cmd))
            uow.tasks.add(new_sprint)
            uow.commit()

    @register(cmd=_commands.StartSprint)
    def start(cmd: _commands.StartSprint,
              handler: Handler,
              context: dict = {}) -> None:
        with handler.uow as uow:
            if not (existing_sprint := uow.tasks.get_by_id(
                    models.sprint.Sprint, cmd.id)):
                raise exceptions.EntityNotFound(
                    f"Sprint id {cmd.id} is not found")
            if existing_sprint.end_date < datetime.today().date():
                raise exceptions.TerkaSprintEndDateInThePast(
                    "Cannot start the sprint, end date in the past")
            if existing_sprint.status.name == "ACTIVE":
                raise exceptions.TerkaSprintActive("Sprint already started")
            if existing_sprint.status.name == "COMPLETED":
                raise exceptions.TerkaSprintCompleted(
                    "Cannot start completed sprint")
            uow.tasks.update(models.sprint.Sprint, cmd.id,
                             {"status": "ACTIVE"})
            for sprint_task in existing_sprint.tasks:
                task = sprint_task.tasks
                task_params = {}
                if task.status.name == "BACKLOG":
                    task_params.update({"status": "TODO"})
                if not task.due_date or task.due_date > existing_sprint.end_date:
                    task_params.update({"due_date": existing_sprint.end_date})
                if task_params:
                    uow.published_events.append(
                        events.TaskUpdated(task.id,
                                           events.UpdateMask(**task_params)))
                if sprint_task.story_points == 0:
                    story_points = input(
                        "Please enter story points estimation "
                        f"for task <{task.id}>: {task.name}: ")
                    try:
                        story_points = float(story_points)
                        uow.published_events.append(
                            events.SprintTaskStoryPointAssigned(
                                sprint_task.id, story_points))
                    except ValueError:
                        print(
                            "[red]Provide number when specifying story points[/red]"
                        )
            uow.commit()
            logging.debug(f"Sprint started, context: {cmd}")

    @register(cmd=_commands.CompleteSprint)
    def complete(cmd: _commands.CompleteSprint,
                 handler: Handler,
                 context: dict = {}) -> None:
        with handler.uow as uow:
            if not (existing_sprint := uow.tasks.get_by_id(
                    models.sprint.Sprint, cmd.id)):
                raise exceptions.EntityNotFound(
                    f"Sprint id {cmd.id} is not found")
            uow.tasks.update(models.sprint.Sprint, cmd.id,
                             {"status": "COMPLETED"})
            for sprint_task in existing_sprint.tasks:
                task = sprint_task.tasks
                task_params = {}
                if task.status.name == "TODO":
                    task_params.update({"status": "BACKLOG"})
                if task.due_date:
                    task_params.update({"due_date": None})
                if task_params:
                    uow.published_events.append(
                        events.TaskUpdated(task.id,
                                           events.UpdateMask(**task_params)))
            uow.commit()
            uow.published_events.append(events.SprintCompleted(cmd.id))
            if comment := cmd.comment:
                uow.published_events.append(
                    events.Sprintommented(id=cmd.id, text=comment))
            uow.commit()
            logging.debug(f"Sprint completed, context: {cmd}")
        handler.publisher.publish("Topic", events.SprintCompleted(cmd.id))

    @register(cmd=_commands.ListSprint)
    def list(cmd: _commands.ListSprint,
             handler: Handler,
             context: dict = {}) -> None:
        with handler.uow as uow:
            if sprints := uow.tasks.list(models.sprint.Sprint):
                handler.printer.console.print_sprint(
                    sprints, printer.PrintOptions.from_kwargs(**context))


class SprintEventHandlers:

    def assign_story_points(event: events.SprintTaskStoryPointAssigned,
                            handler: Handler,
                            context: dict = {}) -> None:
        with handler.uow as uow:
            uow.tasks.update(models.sprint.SprintTask, event.id,
                             {"story_points": event.story_points})
            uow.commit()
            logging.debug(f"Story point assigned, context: {event}")


class TaskCommandHandlers:

    @register(cmd=_commands.CreateTask)
    def create(cmd: _commands.CreateTask,
               handler: Handler,
               context: dict = {}) -> None:
        # TODO: context should be taken from console as well
        if not cmd.name:
            cmd, context = templates.create_command_from_editor(
                models.task.Task, _commands.CreateTask)
        with handler.uow as uow:
            cmd = convert_project(cmd, handler)
            new_task = models.task.Task(**asdict(cmd))
            uow.tasks.add(new_task)
            uow.flush()
            uow.published_events.append(events.TaskCreated(new_task.id))
            uow.commit()
            TaskCommandHandlers._process_extra_args(new_task.id, context, uow)
            handler.publisher.publish("Topic",
                                      events.TaskCompleted(new_task.id))
            handler.printer.console.print_new_object(new_task)

    @register(cmd=_commands.UpdateTask)
    def update(cmd: _commands.UpdateTask,
               handler: Handler,
               context: dict = {}) -> None:
        with handler.uow as uow:
            if not (existing_task := uow.tasks.get_by_id(
                    models.task.Task, cmd.id)):
                raise exceptions.EntityNotFound(
                    f"Task id {cmd.id} is not found")
            if not cmd:
                cmd, context = templates.create_command_from_editor(
                    existing_task, _commands.UpdateTask)
                cmd.id = existing_task.id
            cmd = convert_project(cmd, handler)
            uow.tasks.update(models.task.Task, cmd.id,
                             cmd.get_only_set_attributes())
            uow.commit()
            TaskCommandHandlers._process_extra_args(cmd.id, context, uow)

    @register(cmd=_commands.CollaborateTask)
    def collaborate(cmd: _commands.CollaborateTask,
                    handler: Handler,
                    context: dict = {}) -> None:
        with handler.uow as uow:
            if not (existing_task := uow.tasks.get_by_id(
                    models.task.Task, cmd.id)):
                raise exceptions.EntityNotFound(
                    f"Task id {cmd.id} is not found")
            if not (existing_user := uow.tasks.list(
                    models.user.User, {"name": cmd.collaborator})):
                new_user = models.user.User(name=cmd.collaborator)
                uow.tasks.add(new_user)
                uow.flush()
                user_id = new_user.id
            else:
                user_id = existing_user[0].id
            if not uow.tasks.list(models.collaborator.TaskCollaborator, {
                    "task": cmd.id,
                    "collaborator": user_id
            }):
                uow.tasks.add(
                    models.collaborator.TaskCollaborator(
                        id=cmd.id, collaborator_id=user_id))
                uow.commit()

    @register(cmd=_commands.AddTask)
    def add(cmd: _commands.AddTask,
            handler: Handler,
            context: dict = {}) -> None:

        for field in cmd.__dataclass_fields__:
            if field_value := getattr(cmd, field):
                updated_context = {
                    "entity_type": field,
                    "entity_id": field_value
                }
                if field == "sprint":
                    TaskCommandHandlers._add(
                        _commands.AddTask(cmd.id, sprint=field_value), handler,
                        updated_context)
                if field == "epic":
                    TaskCommandHandlers._add(
                        _commands.AddTask(cmd.id, epic=field_value), handler,
                        updated_context)
                if field == "story":
                    TaskCommandHandlers._add(
                        _commands.AddTask(cmd.id, story=field_value), handler,
                        updated_context)

    def _add(cmd: _commands.AddTask, handler: Handler, context) -> None:
        entity_name, entity_id = context["entity_type"], context["entity_id"]
        entity_module = getattr(models, entity_name)
        entity = getattr(entity_module, entity_name.capitalize())
        entity_task_type = getattr(entity_module,
                                   f"{entity_name.capitalize()}Task")
        entity_dict = {"task": cmd.id}
        entity_dict[entity_name] = entity_id
        with handler.uow as uow:
            if not (existing_entity := uow.tasks.get_by_id(entity, entity_id)):
                raise exceptions.EntityNotFound(
                    f"{entity_name} id {entity_id} is not found")
            if not uow.tasks.list(entity_task_type, entity_dict):
                # raise exceptions.TaskAddedToEntity(
                #     f"task {cmd.id} already added to "
                #     f"{entity_name} {cmd.entity_id}")
                entity_task = entity_task_type(**entity_dict)
                uow.tasks.add(entity_task)
                uow.commit()
                logging.debug(f"Task added to {entity_name}, context {cmd}")
                if entity_name == "sprint":
                    if existing_task := uow.tasks.get_by_id(
                            models.task.Task, cmd.id):
                        task_params = {}
                        if existing_task.status.name == "BACKLOG":
                            task_params.update({"status": "TODO"})
                        if (not existing_task.due_date
                                or existing_task.due_date
                                > existing_entity.end_date
                                or existing_task.due_date
                                < existing_entity.start_date):
                            task_params.update(
                                {"due_date": existing_entity.end_date})
                        if task_params:
                            uow.published_events.append(
                                events.TaskUpdated(
                                    cmd.id, events.UpdateMask(**task_params)))

    @register(cmd=_commands.AssignTask)
    def assign(cmd: _commands.AssignTask,
               handler: Handler,
               context: dict = {}) -> None:
        ...

    @register(cmd=_commands.CompleteTask)
    def complete(cmd: _commands.CompleteTask,
                 handler: Handler,
                 context: dict = {}) -> None:
        with handler.uow as uow:
            uow.tasks.update(models.task.Task, cmd.id, {"status": "DONE"})
            task_completed_event = events.TaskCompleted(cmd.id)
            uow.published_events.append(task_completed_event)
            if comment := cmd.comment:
                uow.published_events.append(
                    events.TaskCommentAdded(id=cmd.id, text=comment))
            if hours := cmd.hours:
                uow.published_events.append(
                    events.TaskHoursSubmitted(id=cmd.id, hours=hours))
            uow.commit()
            handler.publisher.publish("Topic", task_completed_event)

    @register(cmd=_commands.DeleteTask)
    def delete(cmd: _commands.DeleteTask,
               handler: Handler,
               context: dict = {}) -> None:

        for field in cmd.__dataclass_fields__:
            if field_value := getattr(cmd, field):
                updated_context = {
                    "entity_type": field,
                    "entity_id": field_value
                }
                if field == "sprint":
                    TaskCommandHandlers._delete(
                        _commands.DeleteTask(cmd.id, sprint=field_value),
                        handler, updated_context)
                if field == "epic":
                    TaskCommandHandlers._delete(
                        _commands.DeleteTask(cmd.id, epic=field_value),
                        handler, updated_context)
                if field == "story":
                    TaskCommandHandlers._delete(
                        _commands.DeleteTask(cmd.id, story=field_value),
                        handler, updated_context)

    def _delete(cmd: _commands.DeleteTask,
                handler: Handler,
                context: dict = {}) -> None:
        with handler.uow as uow:
            if "entity_type" in context and "entity_id" in context:
                entity_name, entity_id = context["entity_type"], context[
                    "entity_id"]
                entity_module = getattr(models, entity_name)
                entity = getattr(entity_module, entity_name.capitalize())
                entity_task_type = getattr(entity_module,
                                           f"{entity_name.capitalize()}Task")
                entity_dict = {"task": cmd.id}
                entity_dict[entity_name] = entity_id
                if not (existing_entity := uow.tasks.get_by_id(
                        entity, entity_id)):
                    raise exceptions.EntityNotFound(
                        f"{entity_name} id {entity_id} is not found")
                if existing_entity_task := uow.tasks.list(
                        entity_task_type, entity_dict):
                    uow.tasks.delete(entity_task_type, cmd.id)
                    uow.commit()
                    logging.debug(
                        f"Task deleted from {entity_name.capitalize()} "
                        f"{entity_id}, context {cmd}")
                    if entity_name == "sprint":
                        if existing_task := uow.tasks.get_by_id(
                                models.task.Task, cmd.id):
                            task_params = {}
                            if existing_task.status.name == "TODO":
                                task_params.update({"status": "BACKLOG"})
                            task_params.update({"due_date": None})
                            if task_params:
                                uow.published_events.append(
                                    events.TaskUpdated(
                                        cmd.id,
                                        events.UpdateMask(**task_params)))
            else:
                uow.tasks.update(models.task.Task, cmd.id,
                                 {"status": "DELETED"})
                uow.published_events.append(events.TaskDeleted(cmd.id))
            if comment := cmd.comment:
                uow.published_events.append(
                    events.TaskCommentAdded(id=cmd.id, text=comment))
            if hours := cmd.hours:
                uow.published_events.append(
                    events.TaskHoursSubmitted(id=cmd.id, hours=hours))
            uow.commit()
        handler.publisher.publish("Topic", events.TaskCompleted(cmd.id))

    @register(cmd=_commands.CommentTask)
    def comment(cmd: _commands.CommentTask,
                handler: Handler,
                context: dict = {}) -> None:
        with handler.uow as uow:
            uow.tasks.add(
                models.commentary.TaskCommentary(id=cmd.id, text=cmd.text))
            uow.tasks.update(models.task.Task, cmd.id,
                             {"modification_date": datetime.now()})
            uow.commit()

    @register(cmd=_commands.TagTask)
    def tag(cmd: _commands.TagTask,
            handler: Handler,
            context: dict = {}) -> None:
        with handler.uow as uow:
            if not (existing_task := uow.tasks.get_by_id(
                    models.task.Task, cmd.id)):
                raise exceptions.EntityNotFound(
                    f"Task id {cmd.id} is not found")
            if not (existing_tag := uow.tasks.list(models.tag.BaseTag,
                                                   {"text": cmd.tag})):
                new_tag = models.tag.BaseTag(text=cmd.tag)
                uow.tasks.add(new_tag)
                uow.flush()
                tag_id = new_tag.id
            else:
                tag_id = existing_tag[0].id
            if not uow.tasks.list(models.tag.TaskTag, {
                    "task": cmd.id,
                    "tag": tag_id
            }):
                uow.tasks.add(models.tag.TaskTag(id=cmd.id, tag_id=tag_id))
                uow.commit()

    def _process_extra_args(id, context, uow):
        if tags := context.get("tags"):
            for tag in tags.split(","):
                uow.published_events.append(events.TaskTagAdded(id=id,
                                                                tag=tag))
        if collaborators := context.get("collaborators"):
            for collaborator_name in collaborators.split(","):
                uow.published_events.append(
                    events.TaskCollaboratorAdded(
                        id=id, collaborator=collaborator_name))
        if sprints := context.get("sprint"):
            for sprint_id in sprints.split(","):
                uow.published_events.append(
                    events.TaskAddedToSprint(id=id, sprint_id=sprint_id))
        if epics := context.get("epic"):
            for epic_id in epics.split(","):
                uow.published_events.append(
                    events.TaskAddedToEpic(id=id, epic_id=epic_id))
        if stories := context.get("story"):
            for story_id in stories.split(","):
                uow.published_events.append(
                    events.TaskAddedToStory(id=id, story_id=story_id))
        if comment := context.get("comment"):
            uow.published_events.append(
                events.TaskCommentAdded(id=id, text=comment))
        if hours := context.get("time_spent"):
            uow.published_events.append(
                events.TaskHoursSubmitted(id=id, hours=hours))


class TaskEventHandlers(Handler):

    @register(event=events.TaskCreated)
    def created(event: events.TaskCreated,
                handler: Handler,
                context: dict = {}) -> None:
        # TODO: Decide what to do here
        ...

    @register(event=events.TaskCompleted)
    def completed(event: events.TaskCompleted,
                  handler: Handler,
                  context: dict = {}) -> None:
        with handler.uow as uow:
            task_event = models.event_history.TaskEvent(task_id=event.id,
                                                        event_type="STATUS",
                                                        old_value="",
                                                        new_value="DONE")
            uow.tasks.add(task_event)
            uow.commit()

    @register(event=events.TaskUpdated)
    def updated(event: events.TaskUpdated,
                handler: Handler,
                context: dict = {}) -> None:
        with handler.uow as uow:
            uow.tasks.update(models.task.Task, event.id,
                             event.update_mask.get_only_set_attributes())
            current_task = uow.tasks.get_by_id(models.task.Task, event.id)
            for key, updated_value in asdict(event.update_mask).items():
                old_value = getattr(current_task, key)
                if hasattr(old_value, "name"):
                    old_value = old_value.name
                if old_value != updated_value:
                    if updated_value or key.endswith("date"):
                        task_event = models.event_history.TaskEvent(
                            task_id=event.id,
                            event_type=key.upper(),
                            old_value=old_value,
                            new_value=updated_value)
                        uow.tasks.add(task_event)
            uow.commit()
            logging.debug(f"Task updated, context {event}")

    @register(event=events.TaskDeleted)
    def deleted(event: events.TaskDeleted,
                handler: Handler,
                context: dict = {}) -> None:
        with handler.uow as uow:
            task_event = models.event_history.TaskEvent(task_id=event.id,
                                                        event_type="STATUS",
                                                        old_value="",
                                                        new_value="DELETED")
            uow.tasks.add(task_event)
            uow.commit()

    @register(event=events.TaskTagAdded)
    def tag_added(event: events.TaskTagAdded,
                  handler: Handler,
                  context: dict = {}) -> None:
        TaskCommandHandlers.tag(cmd=_commands.TagTask(**asdict(event)),
                                handler=handler,
                                context=context)

    @register(event=events.TaskCommentAdded)
    def comment_added(event: events.TaskCommentAdded,
                      handler: Handler,
                      context: dict = {}) -> None:
        TaskCommandHandlers.comment(cmd=_commands.CommentTask(**asdict(event)),
                                    handler=handler,
                                    context=context)

    @register(event=events.TaskCollaboratorAdded)
    def collaborator_added(event: events.TaskCollaboratorAdded,
                           handler: Handler,
                           context: dict = {}) -> None:
        TaskCommandHandlers.collaborate(
            cmd=_commands.CollaborateTask(**asdict(event)),
            handler=handler,
            context=context)

    @register(event=events.TaskHoursSubmitted)
    def hours_submitted(event: events.TaskHoursSubmitted,
                        handler: Handler,
                        context: dict = {}) -> None:
        with handler.uow as uow:
            uow.tasks.add(
                models.time_tracker.TimeTrackerEntry(
                    task=event.id, time_spent_minutes=event.hours))
            uow.commit()

    @register(event=events.TaskAddedToEpic)
    def added_to_epic(event: events.TaskAddedToEpic,
                      handler: Handler,
                      context: dict = {}) -> None:
        TaskCommandHandlers.add(cmd=_commands.AddTask(id=event.id,
                                                      epic=event.epic_id),
                                handler=handler,
                                context=context)

    @register(event=events.TaskAddedToSprint)
    def added_to_sprint(event: events.TaskAddedToSprint,
                        handler: Handler,
                        context: dict = {}) -> None:
        TaskCommandHandlers.add(cmd=_commands.AddTask(id=event.id,
                                                      sprint=event.sprint_id),
                                handler=handler,
                                context=context)

    @register(event=events.TaskAddedToStory)
    def added_to_story(event: events.TaskAddedToStory,
                       handler: Handler,
                       context: dict = {}) -> None:
        TaskCommandHandlers.add(cmd=_commands.AddTask(id=event.id,
                                                      story=event.story_id),
                                handler=handler,
                                context=context)


class ProjectCommandHandlers:

    @register(cmd=_commands.CreateProject)
    def create(cmd: _commands.CreateProject,
               handler: Handler,
               context: dict = {}) -> None:
        if not cmd.name:
            cmd, context = templates.create_command_from_editor(
                models.project.Project, _commands.CreateProject)
        project_id = None
        with handler.uow as uow:
            if not (existing_project := uow.tasks.get(models.project.Project,
                                                      cmd.name)):
                cmd = convert_workspace(cmd, handler)
                new_project = models.project.Project(**asdict(cmd))
                uow.tasks.add(new_project)
                uow.flush()
                project_id = int(new_project.id)
                new_event = events.ProjectCreated(project_id)
                uow.published_events.append(new_event)
                uow.commit()
                handler.printer.console.print_new_object(new_project)
                handler.publisher.publish("Topic", new_event)
            else:
                logging.warning(f"Project {cmd.name} already exists")
                project_id = existing_project.id
            return project_id

    @register(cmd=_commands.UpdateProject)
    def update(cmd: _commands.UpdateProject,
               handler: Handler,
               context: dict = {}) -> None:
        with handler.uow as uow:
            project = ProjectCommandHandlers._validate_project(cmd.id, uow)
            if not cmd.name:
                cmd, context = templates.create_command_from_editor(
                    project, _commands.UpdateProject)
                cmd.id = project.id
            uow.tasks.update(models.project.Project, project.id,
                             cmd.get_only_set_attributes())
            if tags := context.get("tags"):
                for tag in tags.split(","):
                    uow.published_events.append(
                        events.ProjectTagAdded(id=project.id, tag=tag))
            if collaborators := context.get("collaborators"):
                for collaborator_name in collaborators.split(","):
                    uow.published_events.append(
                        events.ProjecCollaboratorAdded(
                            id=project.id, collaborator=collaborator_name))
            if comment := context.get("comment"):
                uow.published_events.append(
                    events.ProjectCommentAdded(id=project_id, text=comment))
            uow.commit()

    @register(cmd=_commands.CompleteProject)
    def complete(cmd: _commands.CompleteProject,
                 handler: Handler,
                 context: dict = {}) -> None:
        with handler.uow as uow:
            project = ProjectCommandHandlers._validate_project(cmd.id, uow)
            uow.tasks.update(models.project.Project, project.id,
                             {"status": "COMPLETED"})
            uow.published_events.append(events.ProjectCompleted(project.id))
            if comment := cmd.comment:
                uow.published_events.append(
                    events.ProjectCommented(id=project.id, text=comment))
            uow.commit()
        handler.publisher.publish("Topic", events.ProjectCompleted(project.id))

    @register(cmd=_commands.DeleteProject)
    def delete(cmd: _commands.DeleteProject,
               handler: Handler,
               context: dict = {}) -> None:
        with handler.uow as uow:
            project = ProjectCommandHandlers._validate_project(cmd.id, uow)
            uow.tasks.update(models.project.Project, project.id,
                             {"status": "DELETED"})
            uow.published_events.append(events.ProjectDeleted(project.id))
            if comment := cmd.comment:
                uow.published_events.append(
                    events.ProjectCommented(id=project.id, text=comment))
            uow.commit()
            handler.publisher.publish("Topic",
                                      events.ProjectDeleted(project.id))

    @register(cmd=_commands.CommentProject)
    def comment(cmd: _commands.CommentProject,
                handler: Handler,
                context: dict = {}) -> None:
        with handler.uow as uow:
            project = ProjectCommandHandlers._validate_project(cmd.id, uow)
            uow.tasks.add(
                models.commentary.ProjectCommentary(id=project.id,
                                                    text=cmd.text))
            uow.commit()

    @register(cmd=_commands.TagProject)
    def tag(cmd: _commands.TagProject,
            handler: Handler,
            context: dict = {}) -> None:
        with handler.uow as uow:
            project = ProjectCommandHandlers._validate_project(cmd.id, uow)
            if not (existing_tag := uow.tasks.list(models.tag.BaseTag,
                                                   {"text": cmd.tag})):
                new_tag = models.tag.BaseTag(text=cmd.tag)
                uow.tasks.add(new_tag)
                uow.flush()
                tag_id = new_tag.id
            else:
                tag_id = existing_tag[0].id
            if not uow.tasks.list(models.tag.ProjectTag, {
                    "project": project.id,
                    "tag": tag_id
            }):
                uow.tasks.add(
                    models.tag.ProjectTag(id=project.id, tag_id=tag_id))
                uow.commit()

    @register(cmd=_commands.ListProject)
    def list(cmd: _commands.ListProject,
             handler: Handler,
             context: dict = {}) -> None:
        with handler.uow as uow:
            if projects := uow.tasks.list(models.project.Project):
                handler.printer.console.print_project(
                    projects, printer.PrintOptions.from_kwargs(**context))

    def _validate_project(id: str | int, uow) -> models.project.Project:
        if id.isnumeric():
            if not (existing_project := uow.tasks.get_by_id(
                    models.project.Project, id)):
                raise exceptions.EntityNotFound(
                    f"Project id {id} is not found")
        else:
            if not (existing_project := uow.tasks.get(models.project.Project,
                                                      id)):

                raise exceptions.EntityNotFound(
                    f"Project id {id} is not found")
        return existing_project


class ProjectEventHandlers:

    @register(event=events.ProjectCreated)
    def created(event: events.ProjectCreated,
                handler: Handler,
                context: dict = {}) -> None:
        # TODO: Decide what to do here
        ...

    @register(event=events.ProjectCommented)
    def commented(event: events.ProjectCommented,
                  handler: Handler,
                  context: dict = {}) -> None:
        ProjectCommandHandlers.comment(
            cmd=_commands.CommentProject(**asdict(event)), handler=handler)


class EpicCommandHandlers:

    @register(cmd=_commands.CreateEpic)
    def create(cmd: _commands.CreateEpic,
               handler: Handler,
               context: dict = {}):
        if not cmd.name:
            cmd, context = templates.create_command_from_editor(
                models.epic.Epic, type(cmd))
        with handler.uow as uow:
            cmd = convert_project(cmd, handler)
            new_epic = models.epic.Epic(**asdict(cmd))
            uow.tasks.add(new_epic)
            uow.commit()
            handler.printer.console.print_new_object(new_epic)

    @register(cmd=_commands.CompleteEpic)
    def complete(cmd: _commands.CompleteEpic,
                 handler: Handler,
                 context: dict = {}) -> None:
        with handler.uow as uow:
            uow.tasks.update(models.project.Epic, cmd.id,
                             {"status": "COMPLETED"})
            uow.published_events.append(events.EpicCompleted(cmd.id))
            if comment := cmd.comment:
                uow.published_events.append(
                    events.EpicCommented(id=cmd.id, text=comment))
            uow.commit()
        handler.publisher.publish("Topic", events.EpicCompleted(cmd.id))

    @register(cmd=_commands.DeleteEpic)
    def delete(cmd: _commands.DeleteEpic,
               handler: Handler,
               context: dict = {}) -> None:
        with handler.uow as uow:
            uow.tasks.update(models.epic.Epic, cmd.id, {"status": "DELETED"})
            uow.published_events.append(events.EpicDeleted(cmd.id))
            if comment := cmd.comment:
                uow.published_events.append(
                    events.EpicCommented(id=cmd.id, text=comment))
            uow.commit()
        handler.publisher.publish("Topic", events.EpicDeleted(cmd.id))

    @register(cmd=_commands.CommentEpic)
    def comment(cmd: _commands.CommentEpic,
                handler: Handler,
                context: dict = {}) -> None:
        with handler.uow as uow:
            uow.tasks.add(
                models.commentary.EpicCommentary(id=cmd.id, text=cmd.text))
            uow.commit()

    @register(cmd=_commands.AddEpic)
    def add(cmd: _commands.AddEpic,
            handler: Handler,
            context: dict = {}) -> None:
        with handler.uow as uow:
            if not (existing_epic := uow.tasks.get_by_id(
                    models.epic.Epic, cmd.id)):
                raise exceptions.EntityNotFound(f"Epic {cmd.id} is not found")
            if not uow.tasks.get_by_id(models.sprint.Sprint, cmd.sprint_id):
                raise exceptions.EntityNotFound(
                    f"Sprint {cmd.sprint_id} is not found")
            for epic_task in existing_epic.tasks:
                task = epic_task.tasks
                if task.status.name not in ("DONE", "DELETED"):
                    TaskCommandHandlers.add(
                        _commands.AddTask(id=task.id,
                                          entity_type="sprint",
                                          entity_id=cmd.sprint_id), handler)

    @register(cmd=_commands.ListEpic)
    def list(cmd: _commands.ListEpic,
             handler: Handler,
             context: dict = {}) -> None:
        with handler.uow as uow:
            if epics := uow.tasks.list(models.epic.Epic):
                handler.printer.console.print_composite(
                    epics, uow.tasks,
                    printer.PrintOptions.from_kwargs(**context), "epic")


class EpicEventHandlers:
    ...


class StoryCommandHandlers:

    @register(cmd=_commands.CreateStory)
    def create(cmd: _commands.CreateStory,
               handler: Handler,
               context: dict = {}):
        if not cmd.name:
            cmd, context = templates.create_command_from_editor(
                models.story.Story, type(cmd))
        with handler.uow as uow:
            cmd = convert_project(cmd, handler)
            new_story = models.story.Story(**asdict(cmd))
            uow.tasks.add(new_story)
            uow.commit()
            handler.printer.console.print_new_object(new_story)

    @register(cmd=_commands.CompleteStory)
    def complete(cmd: _commands.CompleteStory,
                 handler: Handler,
                 context: dict = {}) -> None:
        with handler.uow as uow:
            uow.tasks.update(models.project.Story, cmd.id,
                             {"status": "COMPLETED"})
            uow.published_events.append(events.StoryCompleted(cmd.id))
            if comment := cmd.comment:
                uow.published_events.append(
                    events.StoryCommented(id=cmd.id, text=comment))
            uow.commit()
        handler.publisher.publish("Topic", events.StoryCompleted(cmd.id))

    @register(cmd=_commands.DeleteStory)
    def delete(cmd: _commands.DeleteStory,
               handler: Handler,
               context: dict = {}) -> None:
        with handler.uow as uow:
            uow.tasks.update(models.story.Story, cmd.id, {"status": "DELETED"})
            uow.published_events.append(events.StoryDeleted(cmd.id))
            if comment := cmd.comment:
                uow.published_events.append(
                    events.StoryCommented(id=cmd.id, text=comment))
            uow.commit()
        handler.publisher.publish("Topic", events.StoryDeleted(cmd.id))

    @register(cmd=_commands.CommentStory)
    def comment(cmd: _commands.CommentStory,
                handler: Handler,
                context: dict = {}) -> None:
        with handler.uow as uow:
            uow.tasks.add(
                models.commentary.StoryCommentary(id=cmd.id, text=cmd.text))
            uow.commit()

    @register(cmd=_commands.AddStory)
    def add(cmd: _commands.AddStory,
            handler: Handler,
            context: dict = {}) -> None:
        with handler.uow as uow:
            if not (existing_story := uow.tasks.get_by_id(
                    models.story.Story, cmd.id)):
                raise exceptions.EntityNotFound(f"Story {cmd.id} is not found")
            if not uow.tasks.get_by_id(models.sprint.Sprint, cmd.sprint_id):
                raise exceptions.EntityNotFound(
                    f"Sprint {cmd.sprint_id} is not found")
            for story_task in existing_story.tasks:
                task = story_task.tasks
                if task.status.name not in ("DONE", "DELETED"):
                    TaskCommandHandlers.add(
                        _commands.AddTask(id=task.id,
                                          entity_type="sprint",
                                          entity_id=cmd.sprint_id), handler)


class WorkspaceCommandHandlers:

    @register(cmd=_commands.CreateWorkspace)
    def create(cmd: _commands.CreateWorkspace,
               handler: Handler,
               context: dict = {}) -> None:
        if not cmd:
            cmd, context = templates.create_command_from_editor(
                models.sprint.Sprint, type(cmd))
        with handler.uow as uow:
            if not uow.tasks.get(models.workspace.Workspace, cmd.name):
                new_workspace = models.workspace.Workspace(**asdict(cmd))
                uow.tasks.add(new_workspace)
                uow.commit()
                handler.printer.console.print_new_object(new_workspace)
            else:
                logging.warning(f"Workspace {cmd.name} already exists")


class TagCommandHandlers:

    @register(cmd=_commands.ListTag)
    def list(cmd: _commands.ListTag,
             handler: Handler,
             context: dict = {}) -> None:
        with handler.uow as uow:
            if tags := uow.tasks.list(models.tag.BaseTag):
                handler.printer.console.print_tag(tags)


class UserCommandHandlers:

    @register(cmd=_commands.ListUser)
    def list(cmd: _commands.ListUser,
             handler: Handler,
             context: dict = {}) -> None:
        with handler.uow as uow:
            if users := uow.tasks.list(models.user.User):
                handler.printer.console.print_user(users)


def convert_project(cmd: _commands.Command,
                    handler: Handler,
                    context: dict = {}) -> Type[_commands.Command]:
    if not (project_name := cmd.project):
        cmd.project = None
        return cmd
    if project_name.isnumeric():
        cmd.project = int(project_name)
        return cmd
    if not (existing_project := handler.uow.tasks.get(models.project.Project,
                                                      project_name)):
        print(f"Creating new project: {project_name}. "
              "Do you want to continue (Y/n)?")
        answer = input()
        while answer.lower() != "y":
            print("Provide a project name: ")
            project_name = input()
            print(f"Creating new project: {project_name}. "
                  "Do you want to continue (Y/n)?")
            answer = input()
        project_id = ProjectCommandHandlers.create(
            cmd=_commands.CreateProject(name=project_name), handler=handler)
        cmd.project = project_id
    else:
        cmd.project = int(existing_project.id)
    return cmd


def convert_workspace(cmd: _commands.Command,
                      handler: Handler,
                      context: dict = {}) -> Type[_commands.Command]:
    if not (workspace := cmd.workspace):
        cmd.workspace = 1
        return cmd
    if workspace.isnumeric():
        cmd.workspace = int(workspace)
        return cmd
    if not (existing_workspace := handler.uow.tasks.get(
            models.workspace.Workspace, workspace)):
        print(f"Creating new workspace: {workspace}. "
              "Do you want to continue (Y/n)?")
        answer = input()
        while answer.lower() != "y":
            print("Provide workspace name: ")
            project_name = input()
            print(f"Creating new workspace: {workspace}. "
                  "Do you want to continue (Y/n)?")
            answer = input()
        workspace_id = WorkspaceCommandHandlers.create(
            cmd=_commands.CreateWorkspace(name=workspace), handler=handler)
        cmd.workspace = workspace_id
    else:
        cmd.workspace = int(existing_workspace.id)
    return cmd



from __future__ import annotations
from dataclasses import dataclass
from datetime import datetime
import inspect
import rich
from rich.console import Console
from rich.table import Table

from terka.service_layer import formatter


@dataclass
class PrintOptions:
    show_tasks: bool = True
    show_history: bool = False
    show_commentaries: bool = False
    show_completed: bool = False
    show_epics: bool = True
    show_stories: bool = True
    show_notes: bool = True
    show_viz: bool = False
    columns: str = ""
    expand_table: bool = True

    @classmethod
    def from_kwargs(cls, **kwargs: dict) -> "PrintOptions":
        return cls(**{k: kwargs[k] for k in kwargs if k in cls.__match_args__})


class Printer:

    def __init__(self):
        self.tui = None
        self.console = ConsolePrinter()


class ConsolePrinter:

    def __init__(self, box=rich.box.SIMPLE, config=None) -> None:
        self.console = Console()
        self.box = box
        self.config = config

    def print_new_object(self, obj, project=None):
        table = Table(box=self.box)
        attributes = self._get_attributes(obj)
        for column, value in attributes:
            if value:
                table.add_column(column)
        table.add_row(*list(zip(*attributes))[1])
        if table.row_count:
            self.console.print(f"Added new {obj.__class__.__name__}")
            self.console.print(table)

    def print_tag(self, entities):
        table = Table(box=self.box)
        for column in ("id", "text"):
            table.add_column(column)
        seen_tags = set()
        for entity in entities:
            text = entity.text
            if text and text not in seen_tags and not text.startswith(
                ("sprint:", "epic:", "story:")):
                table.add_row(f"[red]{entity.id}[/red]", entity.text)
                seen_tags.add(entity.text)
        if table.row_count:
            self.console.print(table)

    def print_user(self, entities):
        table = Table(box=self.box)
        for column in ("id", "name"):
            table.add_column(column)
        seen_users = set()
        for entity in entities:
            name = entity.name
            if name and name not in seen_users:
                table.add_row(f"[red]{entity.id}[/red]", name)
                seen_users.add(name)
        if table.row_count:
            self.console.print(table)

    def print_composite(self,
                        entities,
                        repo,
                        print_options,
                        composite_type,
                        kwargs=None):
        table = Table(box=self.box,
                      title=composite_type.upper(),
                      expand=print_options.expand_table)
        non_active_entities = Table(box=self.box,
                                    title=f"INACTIVE {composite_type.upper()}",
                                    expand=print_options.expand_table)
        default_columns = ("id", "name", "description", "status", "project",
                           "tasks")
        if print_options.columns:
            printable_columns = print_options.columns.split(",")
        else:
            printable_columns = default_columns
        for column in printable_columns:
            table.add_column(column, style="bold")
        for column in printable_columns:
            non_active_entities.add_column(column, style="bold")
        for i, entity in enumerate(entities):
            # TODO: Get project name
            # project = services.lookup_project_name(
            #     entity.project, repo)
            printable_row = {
                "id": f"{entity.id}",
                "name": str(entity.name),
                "description": entity.description,
                "status": entity.status.name,
                "project": f"{entity.project}",
                "tasks": str(len(entity.open_tasks))
            }
            printable_elements = [
                value for key, value in printable_row.items()
                if key in printable_columns
            ]
            if (entity.status.name == "ACTIVE"
                    or len(entity.completed_tasks) < len(entity.tasks)):
                table.add_row(*printable_elements)
            else:
                non_active_entities.add_row(*printable_elements)
        if table.row_count:
            self.console.print(table)
        if non_active_entities.row_count and print_options.show_completed:
            self.console.print(non_active_entities)
        if viz := print_options.show_viz:
            if "time" in viz:
                time_entries = entity.daily_time_entries_hours()
                self._print_time_utilization(time_entries)

    def print_project(self, entities, print_options, kwargs=None):
        table = Table(box=rich.box.SQUARE_DOUBLE_HEAD,
                      expand=print_options.expand_table)
        non_active_projects = Table(box=rich.box.SQUARE_DOUBLE_HEAD,
                                    expand=print_options.expand_table)
        default_columns = ("id", "name", "description", "status", "open_tasks",
                           "overdue", "stale", "backlog", "todo",
                           "in_progress", "review", "done", "median_task_age",
                           "time_spent")

        if print_options.columns:
            printable_columns = print_options.columns.split(",")
        else:
            printable_columns = default_columns
        for column in printable_columns:
            table.add_column(column)
        for column in printable_columns:
            if column in ("id", "name", "description", "status", "open_tasks"):
                non_active_projects.add_column(column)
        for entity in entities:
            if len(open_tasks :=
                   entity.open_tasks) > 0 and entity.status.name == "ACTIVE":

                if overdue_tasks := entity.overdue_tasks:
                    entity_id = f"[red]{entity.id}[/red]"
                else:
                    entity_id = f"[green]{entity.id}[/green]"
                printable_row = {
                    "id":
                    f"{entity.id}",
                    "name":
                    str(entity.name),
                    "description":
                    entity.description,
                    "status":
                    entity.status.name,
                    "open_tasks":
                    str(len(open_tasks)),
                    "overdue":
                    str(len(overdue_tasks)),
                    "stale":
                    str(len(entity.stale_tasks)),
                    "backlog":
                    str(len(entity.backlog)),
                    "todo":
                    str(len(entity.todo)),
                    "in_progress":
                    str(len(entity.in_progress)),
                    "review":
                    str(len(entity.review)),
                    "done":
                    str(len(entity.done)),
                    "median_task_age":
                    str(entity.median_task_age),
                    "time_spent":
                    formatter.Formatter.format_time_spent(
                        entity.total_time_spent),
                }
                printable_elements = [
                    value for key, value in printable_row.items()
                    if key in printable_columns
                ]
                table.add_row(*printable_elements)
            else:
                printable_row = {
                    "id": f"{entity.id}",
                    "name": str(entity.name),
                    "description": entity.description,
                    "status": entity.status.name,
                    "open_tasks": str(len(open_tasks)),
                }
                printable_elements = [
                    value for key, value in printable_row.items()
                    if key in printable_columns
                ]
                non_active_projects.add_row(*printable_elements)

        if table.row_count:
            self.console.print(table)
        if non_active_projects.row_count and print_options.show_completed:
            self.console.print("[green]Inactive projects[/green]")
            self.console.print(non_active_projects)

    def print_sprint(self, entities, print_options, kwargs=None):
        table = Table(box=rich.box.SQUARE_DOUBLE_HEAD,
                      expand=print_options.expand_table)
        all_sprints = Table(box=rich.box.SQUARE_DOUBLE_HEAD,
                            expand=print_options.expand_table)
        columns = ("id", "start_date", "end_date", "goal", "status",
                   "open tasks", "pct_completed", "velocity", "collaborators",
                   "time_spent", "utilization")
        if print_options.columns:
            printable_columns = print_options.columns.split(",")
        else:
            printable_columns = columns

        for column in printable_columns:
            table.add_column(column)
            all_sprints.add_column(column)
        for entity in entities:
            printable_row = {
                "id":
                str(entity.id),
                "start_date":
                str(entity.start_date),
                "end_date":
                str(entity.end_date),
                "goal":
                entity.goal,
                "status":
                entity.status.name,
                "open tasks":
                f"{len(entity.open_tasks)} ({len(entity.tasks)})",
                "pct_completed":
                f"{round(entity.pct_completed * 100, 2)}%",
                "velocity":
                str(round(entity.velocity, 2)),
                "collaborators":
                entity.collaborators_as_string,
                "time_spent":
                str(
                    formatter.Formatter.format_time_spent(
                        entity.total_time_spent)),
                "utilization":
                f"{round(entity.utilization * 100)}%"
            }
            printable_elements = [
                value for key, value in printable_row.items()
                if key in printable_columns
            ]
            if entity.status.name in ("ACTIVE", "PLANNED"):
                table.add_row(*printable_elements)
            all_sprints.add_row(*printable_elements)

        if print_options.show_completed:
            self.console.print(all_sprints)
        elif table.row_count:
            self.console.print(table)

    def _get_attributes(self, obj) -> list[tuple[str, str]]:
        import inspect
        attributes = []
        for name, value in inspect.getmembers(obj):
            if not name.startswith("_") and not inspect.ismethod(value):
                if not value:
                    continue
                elif name in ("created_by", "assignee"):
                    attributes.append(
                        # (name, services.lookup_user_name(value, self.repo)))
                        (name, str(value)))
                elif name == "project":
                    # project = services.lookup_project_name(value, self.repo)
                    attributes.append((name, str(value)))
                elif hasattr(value, "name"):
                    attributes.append((name, value.name))
                elif isinstance(value, datetime):
                    attributes.append((name, value.strftime("%Y-%m-%d %H:%M")))
                elif isinstance(value, set):
                    values = value.pop()
                    attributes.append((name, str(values)))
                else:
                    attributes.append((name, str(value)))
        return attributes

    def _print_time_utilization(self, time_entries) -> None:
        plt.date_form('Y-m-d')
        plt.plot_size(100, 15)
        plt.title(
            f"Time tracker - {format_hour_minute(sum(time_entries.values()))} spent"
        )
        plt.bar(time_entries.keys(), time_entries.values())
        plt.show()



from typing import Dict, Optional

import abc
from datetime import datetime, timedelta
from sqlalchemy import and_, or_, not_

from terka.domain.element import Element
from terka.domain.task import Task
from terka.domain.project import Project
from terka.domain.user import User
from terka.domain.event_history import TaskEvent
from . import orm


class AbsRepository(abc.ABC):

    def __init__(self):
        pass

    def add(self, element: Element) -> None:
        self._add(element)

    def get(self, element: Element, element_name: str) -> Element:
        element = self._get(element, element_name)
        return element

    def get_by_id(self, element: Element, element_id: int) -> Element:
        element = self._get_by_element_id(element, element_id)
        return element

    @abc.abstractmethod
    def _add(self, element: Element) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def _get(self, element: Element, element_name: str) -> Element:
        raise NotImplementedError @ abc.abstractmethod

    def _get_by_element_id(self, element_type: str,
                           element_id: int) -> Element:
        raise NotImplementedError


class SqlAlchemyRepository(AbsRepository):

    def __init__(self, session):
        super().__init__()
        self.session = session

    def delete(self, element: Element, element_id: str):
        return self.session.query(element).filter_by(id=element_id).delete()

    def update(self, element: Element, element_id: str,
               update_dict: Dict[str, str]):
        return self.session.query(element).filter_by(
            id=element_id).update(update_dict)

    def list(self, element: Element, filter_dict: Optional[Dict[str, str]] = {}):
        query_object = self.session.query(element)
        overdue_check = False
        stale_check = False
        if "overdue" in filter_dict:
            overdue_check = True
            del filter_dict["overdue"]
        if "stale" in filter_dict:
            stale_check = True
            stale_lookback = int(filter_dict["stale"])
            del filter_dict["stale"]
        if filter_dict and isinstance(filter_dict, dict):
            query_dict = {}
            or_values = {}
            for k, v in filter_dict.items():
                if isinstance(v, str):
                    if len((multiple_values := v.split(","))) == 1:
                        query_dict[k] = str(v)
                    else:
                        or_values[k] = [str(v) for v in multiple_values]
                elif isinstance(v, int):
                    query_dict[k] = str(v)
                elif isinstance(v, datetime):
                    query_dict[k] = v
                else:
                    if len(v) > 1:
                        or_values[k] = [str(v) for v in v]
                    else:
                        query_dict[k] = str(v[0])
            if or_values:
                for key, values in or_values.items():
                    if values[0].startswith("NOT"):
                        values = [
                            value.replace("NOT:", "") for value in values
                        ]
                        query_object = query_object.filter(
                            ~getattr(element, key).in_(values))
                    else:
                        query_object = query_object.filter(
                            getattr(element, key).in_(values))
            if query_dict:
                for key, value in query_dict.items():
                    if isinstance(value, str) and value.startswith("NOT"):
                        query_object = query_object.filter(
                            getattr(element, key) != value.replace("NOT:", ""))
                    if isinstance(value, datetime):
                        query_object = query_object.filter(
                            getattr(element, key) <= value)
                    else:
                        query_object = query_object.filter(
                            getattr(element, key) == value)

            if stale_check:
                days_ago = datetime.now().date() - timedelta(days=stale_lookback)
                query_object = query_object.filter(
                    getattr(element,
                            "status").in_(["TODO", "IN_PROGRESS", "REVIEW"])).join(
                                TaskEvent, (element.id == TaskEvent.task)).filter(
                                    TaskEvent.date <= days_ago, TaskEvent.type == "STATUS")
                return query_object.all()
            if overdue_check:
                query_object = query_object.filter(
                    getattr(element, "due_date") < datetime.now().date())
                return query_object.all()
        elif isinstance(filter_dict, str):
            if overdue_check:
                query_object = query_object.filter(
                    getattr(element, "due_date") < datetime.now().date())
            if filter_dict.startswith("NOT:"):
                return query_object.filter(
                    getattr(element, "name") != filter_dict.replace(
                        "NOT:", "")).all()
            else:
                return query_object.filter_by(name=filter_dict).first()
        if overdue_check:
            query_object = query_object.filter(
                getattr(element, "due_date") < datetime.now().date())
        if stale_check:
            days_ago = datetime.now().date() - timedelta(days=stale_lookback)
            query_object = query_object.filter(
                getattr(element,
                        "status").in_(["TODO", "IN_PROGRESS", "REVIEW"])).join(
                            Event, (element.id == Event.element_id)).filter(
                                Event.date <= days_ago, Event.type == "status")
        return query_object.all()

    def _add(self, element):
        self.session.add(element)

    def _get(self, element, element_name):
        return self.session.query(element).filter_by(name=element_name).one_or_none()

    def _get_by_element_id(self, element, element_id):
        return (self.session.query(element).filter_by(id=element_id).one_or_none())
